#! /usr/bin/env python

import logging, logging.config, os, raven
from celery import Celery
from celery.signals import setup_logging
from django.conf import settings

from ._version import get_versions
__version__ = get_versions ()['version']
del get_versions

LOG = logging.getLogger (__name__)
DEFAULT_LOGCONF = "/etc/qeventlog/logging.conf"

os.environ.setdefault ("DJANGO_SETTINGS_MODULE", "qeventlog.settings")
app = Celery ("qeventlog", include = ["qeventlog.tasks",])
app.config_from_object ("django.conf:settings")
app.autodiscover_tasks (lambda: settings.INSTALLED_APPS)
app_name = "qeventlog"
app_ver = __version__
sentry_dsn = settings.RAVEN_CONFIG["dsn"]
SENTRY = raven.Client (sentry_dsn,
                       auto_log_stacks = True,
                       release = "%s: %s" % (app_name, app_ver),
                       transport = raven.transport.http.HTTPTransport)
app.set_current ()

@setup_logging.connect
# @logtool.log_call
def setup_logging_handler (**kwargs): # pylint: disable=W0613
  logging.config.fileConfig (DEFAULT_LOGCONF,
                               disable_existing_loggers = False)
  raven.contrib.celery.register_logger_signal (SENTRY)
  raven.contrib.celery.register_signal (SENTRY)

if __name__ == "__main__":
  app.start()



#! /usr/bin/env python

import architect, datetime, logging, logtool, retryp, uuid
from django.db import models, transaction
from django.contrib.postgres.fields import JSONField
from model_utils import Choices
from model_utils.fields import StatusField

LOG = logging.getLogger (__name__)
DEFAULT_RETRY = 20
TASKNAME_LEN = 128

@architect.install ("partition", type = "range", subtype = "date",
                    constraint = "month", column = "created")
class QEvent (models.Model):
  created = models.DateTimeField (db_index = True)
  timestamp = models.DecimalField (
    max_digits = 30, decimal_places = 6, null = True, blank = True,
    db_index = True)
  event = models.CharField (max_length = 64, db_index = True)
  task_id = models.UUIDField (db_index = True)
  data = JSONField (db_index = True, null = True, blank = True)

  # Because there's not a clear way to have pre and post-save hooks on
  # save() execute in the same transaction.
  @retryp.retryp (count = 5, delay = 0, expose_last_exc = True)
  @logtool.log_call
  @classmethod
  def record (cls, time_t, **kwargs): # pylint: disable=unused-argument
    with transaction.atomic ():
      now = datetime.datetime.utcnow ()
      QEvent (created = now,
              timestamp = time_t,
              event = kwargs["event"],
              task_id = kwargs["uuid"],
              data = kwargs,
            ).save ()
      if kwargs["event"] == "after_task_publish" and kwargs.get ("parent_id"):
        QChildTask (created = now,
                    timestamp = kwargs["timestamp"],
                    parent = kwargs.get ("parent_id"),
                    child = kwargs.get ("uuid")).save ()
      QTaskState.record (now, **kwargs)

  class Meta: # pylint: disable=no-init,too-few-public-methods,old-style-class
    ordering = ["created",]
    index_together = ("event", "task_id", "timestamp")

@architect.install ("partition", type = "range", subtype = "date",
                    constraint = "month", column = "created")
class QChildTask (models.Model):
  created = models.DateTimeField (db_index = True)
  timestamp = models.DecimalField (
    max_digits = 30, decimal_places = 6, null = True, blank = True,
    db_index = True)
  parent = models.UUIDField (db_index = True)
  child = models.UUIDField (db_index = True)

  class Meta: # pylint: disable=no-init,too-few-public-methods,old-style-class
    ordering = ["created", "timestamp"]
    index_together = ("parent", "child")

class QTaskName (models.Model):
  name = models.CharField (unique = True, null = True, blank = True,
                           max_length = TASKNAME_LEN, db_index = True)

  class Meta: # pylint: disable=no-init,too-few-public-methods,old-style-class
    ordering = ["name",]

@architect.install ("partition", type = "range", subtype = "date",
                    constraint = "month", column = "created")
class QTaskState (models.Model):
  # Sorted by their expected order of arrival
  _task_states = ["before_task_publish", "after_task_publish",
                  "task_prerun", "task_retry", "task_postrun",
                  "task_success", "task_failure", "task_revoked",]
  STATUS = Choices (*_task_states)
  task_id = models.UUIDField (db_index = True, primary_key = True)
  created = models.DateTimeField (db_index = True)
  timestamp = models.DecimalField (
    max_digits = 30, decimal_places = 6, null = True, blank = True,
    db_index = True)
  name = models.ForeignKey (QTaskName)
  retries = models.IntegerField (db_index = True, null = True, blank = True)
  status = StatusField (db_index = True)

  @logtool.log_call
  @classmethod
  def record (cls, now, **kwargs):
    name, created = ( # pylint: disable=W0612
            QTaskName.objects.get_or_create ( # pylint: disable=no-member
            name = kwargs["task"]))
    o, created = QTaskState.objects.get_or_create ( # pylint: disable=no-member
      task_id = uuid.UUID (kwargs["uuid"]),
      defaults = {
        "created": now,
        "timestamp": kwargs["timestamp"],
        "name": name,
        "retries": kwargs.get ("retries", 0),
        "status": kwargs["event"],
      })
    if created or ((kwargs.get ("retries", 0) > o.retries)
                   or (cls._task_states.index (kwargs["event"])
                       > cls._task_states.index (o.status))):
      o.timestamp = kwargs["timestamp"]
      o.retries = kwargs.get ("retries", 0)
      o.status = kwargs["event"]
      o.save ()

  class Meta: # pylint: disable=no-init,too-few-public-methods,old-style-class
    ordering = ["created",]
    index_together = ("created", "timestamp", "status")



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-05-03 03:52
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('qeventlog', '0002_qtaskstate_task'),
    ]

    operations = [
        migrations.CreateModel(
            name='QTaskName',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, db_index=True, max_length=128, null=True, unique=True)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.RemoveField(
            model_name='qtaskstate',
            name='task',
        ),
        migrations.AddField(
            model_name='qtaskstate',
            name='name',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, to='qeventlog.QTaskName'),
            preserve_default=False,
        ),
    ]



#! /usr/bin/env python

from __future__ import absolute_import
from celery import Task
from . import settings

class QETask (Task): # pylint: disable=abstract-method

  #: If :const:`True` the task is an abstract base class.
  #: Note: Also set this in derived clases?
  abstract = True

  #: The name of a serializer that are registered with
  #: :mod:kombu.serialization.registry.  Default is 'pickle'.
  serializer = "json"

  #: Maximum number of retries before giving up.  If set to :const:`None`,
  #: it will **never** stop retrying.
  max_retries = settings.FAIL_RETRYCOUNT

  #: Default time to wait and the base multiplier for exponential
  #: backoff on retries.
  default_retry_period = settings.FAIL_WAITTIME

  #: Default time in seconds before a retry of the task should be
  #: executed.  3 minutes by default.
  # default_retry_delay = 3 * 60
  @property
  def default_retry_delay (self):
    return self.default_retry_period * (self.request.retries + 1)

  #: Hard time limit.
  #: Defaults to the :setting:`task_time_limit` setting.
  time_limit = settings.CELERYD_TASK_TIME_LIMIT

  #: Soft time limit.
  #: Defaults to the :setting:`task_soft_time_limit` setting.
  soft_time_limit = settings.CELERYD_TASK_SOFT_TIME_LIMIT

  #: If enabled the task will report its status as 'started' when the task
  #: is executed by a worker.  Disabled by default as the normal behaviour
  #: is to not report that level of granularity.  Tasks are either pending,
  #: finished, or waiting to be retried.
  #:
  #: Having a 'started' status can be useful for when there are long
  #: running tasks and there is a need to report which task is currently
  #: running.
  #:
  #: The application default can be overridden using the
  #: :setting:`task_track_started` setting.
  track_started = True

  #: Even if :attr:`acks_late` is enabled, the worker will
  #: acknowledge tasks when the worker process executing them abrubtly
  #: exits or is signaled (e.g. :sig:`KILL`/:sig:`INT`, etc).
  #:
  #: Setting this to true allows the message to be requeued instead,
  #: so that the task will execute again by the same worker, or another
  #: worker.
  #:
  #: Warning: Enabling this can cause message loops; make sure you know
  #: what you're doing.
  reject_on_worker_lost = True



#!/usr/bin/env python
import os
import sys

if __name__ == "__main__":
  os.environ.setdefault("DJANGO_SETTINGS_MODULE", "qeventlog.settings")
  from django.core.management import execute_from_command_line
  execute_from_command_line(sys.argv)
